package suicideChess;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Random;

import suicideChess.Board.NoPieceOnSquare;
import suicideChess.Move.NotAValidMoveException;
import suicideChess.Square.NotAValidSquare;

/**
 * This class will contain all the AI.
 * @author Jean-Baptiste H&eacute;tier
 * @version $LastChangedRevision$, $LastChangedDate$
 */

public class ComputerPlayer {

    //best move found
    private static Move bestMove = null;
    
    //to tell the number of nodes searched
    private static int nodesSearched;
    
 
    /**
     * This asks the computer to compute a move
     * @param bitboard The current status of the {@link Board}
     * @return move A {@link Move}
     * @throws NotAValidSquare 
     * @see Board
     * @see Move
     */
    public static Move doRandomMove(Board bitboard) throws NotAValidSquare {
        Random generator = new Random();
        Rules.legalMovesForPlayer(bitboard);
        ArrayList<Move> allLegalMoves = Rules.getLegalMovesCapture();
        if (allLegalMoves.size()==0) {
            allLegalMoves = Rules.getLegalMovesNonCapture();
        }
        if (allLegalMoves.size()!=0) {
            return allLegalMoves.get(generator.nextInt(allLegalMoves.size()));
        }
        
        throw new RuntimeException("**Error** in doRandomMove");

    }
    
    /**
     * Basic MinMax
     * @param bitboard The bitboard
     * @return The best Move found 
     * @throws NotAValidSquare 
     * @throws NoPieceOnSquare 
     * @see Board
     * @see Move
     */
    public static Move doMinMaxMove(Board bitboard) throws NotAValidSquare, NoPieceOnSquare {
        bestMove = null;
        nodesSearched = 0;

        int bestScore = MinMax(bitboard, 0);
        if (SuicideChess.postThinkingOutput()) {
            System.out.println(SuicideChess.getPlyDepth()+" "+bestScore*100+" 0 "+nodesSearched+" "+bestMove);
        }
        return bestMove;
    }
    
    
    private static int MinMax(Board bitboard, int currentDepth) throws NotAValidSquare, NoPieceOnSquare {
        nodesSearched++;
        
        if (currentDepth >= SuicideChess.getPlyDepth()) {
            return bitboard.getBoardValue();
        }
        
        Rules.legalMovesForPlayer(bitboard);
        ArrayList<Move> allLegalMoves = Rules.getLegalMovesCapture();
        if (allLegalMoves.size()==0) {
            allLegalMoves = Rules.getLegalMovesNonCapture();
        }
        if (allLegalMoves.size()==0) {
            if (bitboard.getCurrentPlayer()==Piece.BLACK) {
                return Board.BLACK_WINS;
            } else {
                return Board.WHITE_WINS;
            }
        } else {
            ArrayList<Integer> bestMoveIndex=new ArrayList<Integer>();
            int currentScore;
            int bestScoreSoFar;
            if (bitboard.getCurrentPlayer()==Piece.BLACK) {            
                bestScoreSoFar=Board.WHITE_WINS+1; //any move even a WHITE_WINS will be better than that
                for (int i=0; i<allLegalMoves.size(); i++) {
                    Board boardCopy = new Board(bitboard);
                    boardCopy.doMove(allLegalMoves.get(i));
                    currentScore=MinMax(boardCopy,currentDepth+1);
                    if (currentScore <= bestScoreSoFar) { //black tries to minimise his score
                        if (currentScore<bestScoreSoFar) {
                            bestScoreSoFar = currentScore;
                            bestMoveIndex.clear();
                        }
                        bestMoveIndex.add(i);
                    }
                }
            } else { //white piece
                bestScoreSoFar=Board.BLACK_WINS-1; //any move even a BLACK_WINS will be better than that
                for (int i=0; i<allLegalMoves.size(); i++) {
                    Board boardCopy = new Board(bitboard);
                    boardCopy.doMove(allLegalMoves.get(i));
                    currentScore=MinMax(boardCopy,currentDepth+1);
                    if (currentScore >= bestScoreSoFar) { //white tries to maximise his score
                        if (currentScore>bestScoreSoFar) {
                            bestScoreSoFar = currentScore;
                            bestMoveIndex.clear();
                        }
                        bestMoveIndex.add(i);
                    }
                }
            }
            
            //select one of the best moves randomly
            Random generator = new Random();
            if (currentDepth == 0) { System.out.println(bestMoveIndex.size()); }
            bestMove = allLegalMoves.get(bestMoveIndex.get(generator.nextInt(bestMoveIndex.size())));
            
            return bestScoreSoFar;
        }  
    }
   
    
    //variables used in aplha beta pruning
    
    private static boolean quiescenceSearch; //this will be used to determine is quiescence search is needed. 
    private static boolean adaptativeDepth; //this will 
    private static String extraSymbol=""; //display an extra symbol after the depth in the output if doing either quiescence search or adaptative depth
    
    private static Date thinkingBeginingTime;
    private static int maxDepth;
    private static ArrayList<Move> bestMoves=new ArrayList<Move>();
    //private static Move[] killerMoves = new Move[SuicideChess.KILLER_SIZE]; //killer 
    private static Move[] principalVariation;
    
    /**
     * Alpha-Beta
     * @param bitboard The bitboard
     * @return The best Move found 
     * @throws NotAValidSquare 
     * @throws NoPieceOnSquare 
     * @throws NotAValidMoveException 
     * @see Board
     * @see Move
     */
    public static Move doAlphaBetaMove(Board bitboard) throws NotAValidSquare, NoPieceOnSquare, NotAValidMoveException {
        bestMove = null;
        nodesSearched = 0;

        thinkingBeginingTime = new Date();
        quiescenceSearch = false;
        adaptativeDepth = false;
        extraSymbol = "";
        principalVariation=new Move[0];
        
        //iterative deepening
        for(maxDepth=SuicideChess.MIN_PLY_DEPTH; maxDepth<=SuicideChess.getPlyDepth(); maxDepth++) {
            if(maxDepth==SuicideChess.getPlyDepth()) {
               if(SuicideChess.QUIESCENCE_SEARCH) {
                   quiescenceSearch = true; //don't do quiescence search till the last level of iterative deepening
                   extraSymbol = "+"; //display extra symbol in the thinking output
               }
               if(SuicideChess.ADAPTATIVE_DEPTH) {
                   adaptativeDepth = true; //don't do adaptative search till the last level of iterative deepening
                   extraSymbol = "+"; //display extra symbol in the thinking output
               }
            }
            
            //thinkingBeginingTime = new Date();
            //nodesSearched=0;
            //alpha beta pruning.
            ReturnWrapper bestScore = AlphaBeta(bitboard, 0, SuicideChess.PRINCIPAL_VARIATION_FIRST
                    , maxDepth, Board.MIN_VALUE, Board.MAX_VALUE);
            
            Date thinkingEndTime = new Date();
            
            //select one of the best moves randomly
            Random generator = new Random();
            //if (currentDepth == 0) { System.out.println(bestMoveIndex.size()); }
            bestMove = bestMoves.get(generator.nextInt(bestMoves.size()));
            
            if (SuicideChess.postThinkingOutput()) {
                System.out.println(maxDepth+extraSymbol+"\t"+bestScore.getBranchValue()+
                        "\t"+((int)(thinkingEndTime.getTime()-thinkingBeginingTime.getTime())/10)+ //search time in centiseconds
                        "\t"+nodesSearched+"\t"+bestScore.getPrincipalVariation());
            }
            
            //if found a win for black or white.
            if((bitboard.getCurrentPlayer()==Piece.BLACK 
                    && bestScore.getBranchValue()==Board.BLACK_WINS)
                    || (bitboard.getCurrentPlayer()==Piece.WHITE 
                    && bestScore.getBranchValue()==Board.WHITE_WINS)) {
                break; //no need to continue iterative deepening.
            }
            
            if(SuicideChess.PRINCIPAL_VARIATION_FIRST) {
                //parsing the principal variation to be able to analyse it first
                String principalVariationString[] = bestScore.getPrincipalVariation().split("\\s");
                principalVariation = new Move[principalVariationString.length];
                Board playing; //need to update bitboard to be able to generate moves
                playing = new Board(bitboard);
                for (int i = 0; i < principalVariationString.length; i++) {
                    //playing.display();
                    //System.out.println(principalVariationString[i]);
                    principalVariation[i] = new Move(principalVariationString[i],playing);
                    playing.doMove(principalVariation[i]);
                }
            }
        
        }
        
        if(SuicideChess.playInACSII()) {
            System.out.println("Found "+bestMoves.size()+" good moves.");
        }
        
        //bestMove.display();
        
        return bestMove;
    }
    
    
    //this class is used to return two arguments in the next function, the two arguments being
    //an integer representing the value of alpha or beta
    //an integer representing the real value of the branch (alpha-beta only give boundaries)
    private static class ReturnWrapper {
        private int alphaBeta;
        private int branchValue;
        private String principalVariation;
        public ReturnWrapper(int alphaBeta, int branchValue, String principalVariation) {
            this.alphaBeta = alphaBeta;
            this.branchValue = branchValue;
            this.principalVariation = principalVariation;
        }
        public int getAlphaBeta() {return this.alphaBeta;}
        public int getBranchValue() {return this.branchValue;}
        public String getPrincipalVariation() {return this.principalVariation;}
    };
    
    //Alpha beta.
    private static ReturnWrapper AlphaBeta(Board bitboard, int currentDepth, boolean inPrincipalVariation, int currentMaxDepth, int alpha, int beta) throws NotAValidSquare, NoPieceOnSquare {
        nodesSearched++;
        
        if(bitboard.isADraw()) {
            return new ReturnWrapper(Board.DRAW_BOARD,Board.DRAW_BOARD,"");
        }

        if((!quiescenceSearch)&& (currentDepth >= currentMaxDepth)) { //stop if we are getting to deep
            //System.out.println("'-> Evaluate: "+bitboard.getBoardValue());
            return new ReturnWrapper(bitboard.getBoardValue(),bitboard.getBoardValue(),"");
        }
        
        //generate legal moves.
        Rules.legalMovesForPlayer(bitboard);
        ArrayList<Move> allLegalMoves = Rules.getLegalMovesCapture();
        if (allLegalMoves.size()==0) { //if no capture moves
            if((quiescenceSearch) && (currentDepth >= currentMaxDepth)) {
                //System.out.println("'-> Evaluate: "+bitboard.getBoardValue());
                return new ReturnWrapper(bitboard.getBoardValue(),bitboard.getBoardValue(),""); //position is quiet
            }
            allLegalMoves = Rules.getLegalMovesNonCapture();
        } else {
            boolean changedMaxDepth = false; //to make sure that quiescence and adaptative don't interfere
            if((quiescenceSearch) && (currentDepth >= currentMaxDepth)) {
                if((currentMaxDepth<SuicideChess.MAX_QUIESCENCE_DEPTH) && (allLegalMoves.size()<=SuicideChess.QUIESCENCE_LIMIT)) {
                    currentMaxDepth++; //seek a quiet position
                    changedMaxDepth = true;
                } else {
                    return new ReturnWrapper(bitboard.getBoardValue(),bitboard.getBoardValue(),"");
                }
            }
            if (adaptativeDepth) {//if there are captures, see if we can just abandon search here or go further
                if(currentDepth==0 && allLegalMoves.size()==1) {
                    bestMoves.clear();
                    bestMoves.add(allLegalMoves.get(0));
                    Board boardCopy = new Board(bitboard);
                    boardCopy.doMove(allLegalMoves.get(0));
                    return new ReturnWrapper(bitboard.getBoardValue(),bitboard.getBoardValue(),allLegalMoves.get(0).toString());
                }
                if(!changedMaxDepth && allLegalMoves.size()<=SuicideChess.ADAPTATIVE_BRANCHING_LIMIT)
                    currentMaxDepth++; //go one step further in depth
            }
        }
        if (allLegalMoves.size()==0) { //if no moves at all
            if (bitboard.getCurrentPlayer()==Piece.BLACK) {
                //System.out.println("'-> Evaluate *BLACK WINS*: "+Board.BLACK_WINS);
                return new ReturnWrapper(Board.BLACK_WINS,Board.BLACK_WINS,"");
            } else {
                //System.out.println("'-> Evaluate *WHITE WINS* : "+Board.WHITE_WINS);
                return new ReturnWrapper(Board.WHITE_WINS,Board.WHITE_WINS,"");
            }
        } else {
            int currentScore;
            int bestScoreSoFar;
            int currentAlphaBeta;
            String bestVariationSoFar="";
            ReturnWrapper returnValue;
            if (SuicideChess.MOVE_ORDERING) {
                /* DEBUG
                if(currentDepth==0) {
                    String debug = "";
                    for (int i=0; i<allLegalMoves.size(); i++) {
                        Board bitboardCopy = new Board(bitboard);
                        bitboardCopy.doMove(allLegalMoves.get(i));
                        debug += allLegalMoves.get(i).toString()+bitboardCopy.getBoardValue()+" ";
                    }
                    System.out.println(debug);
                }
                */
                
                Collections.sort(allLegalMoves,new MoveCompare(bitboard));
                
                /* DEBUG
                if(currentDepth==0) {
                    String debug = "";
                    for (int i=0; i<allLegalMoves.size(); i++) {
                        Board bitboardCopy = new Board(bitboard);
                        bitboardCopy.doMove(allLegalMoves.get(i));
                        debug += allLegalMoves.get(i).toString()+bitboardCopy.getBoardValue()+" ";
                    }
                    System.out.println(debug);
                }
                */
            }
            if (bitboard.getCurrentPlayer()==Piece.BLACK) {
                bestScoreSoFar=Board.MAX_VALUE; //black tries to minimise
                for (int i=-1; i<allLegalMoves.size(); i++) {
                    Board boardCopy = new Board(bitboard);
                    if (i==-1) { //do the principal variation first
                        if(!inPrincipalVariation || (principalVariation.length <= currentDepth)) {
                            continue;
                        } else {
                            boardCopy.doMove(principalVariation[currentDepth]);
                            //System.out.println(i+"bdoingMove"+currentDepth+" "+principalVariation[currentDepth]);
                            returnValue = AlphaBeta(boardCopy,currentDepth+1,true,currentMaxDepth,Board.MIN_VALUE,beta);
                        }
                    } else { //and then don't do it again.
                        if(inPrincipalVariation && (principalVariation.length > currentDepth)) {
                            if(principalVariation[currentDepth].isSimpleEqualTo(allLegalMoves.get(i))) {
                                continue;
                            }
                        }
                        boardCopy.doMove(allLegalMoves.get(i));
                        //System.out.println(i+"bdoingMove"+currentDepth+" "+allLegalMoves.get(i));
                        returnValue = AlphaBeta(boardCopy,currentDepth+1,false,currentMaxDepth,Board.MIN_VALUE,beta);
                    }
                    //System.out.println("Analysing "+currentDepth+":"+allLegalMoves.get(i));
                    
                    
                    currentScore = returnValue.getBranchValue();
                    //if(currentDepth==0 || currentDepth==1) System.out.println(returnValue.getPrincipalVariation()+" "+currentScore);
                    currentAlphaBeta = returnValue.getAlphaBeta();
                    
                    //System.out.println("| CurrentScore, BestScore:" + currentScore + ", " + bestScoreSoFar);
                    //System.out.println("| CurrentBeta, beta:" + currentAlphaBeta + ", " + beta);
                    //if(returnValue.alphaBeta!=returnValue.branchValue)
                    //    System.out.println(returnValue.alphaBeta+"\t"+returnValue.branchValue);
                    
                    //calculating new value of beta
                    if (currentAlphaBeta<beta) {
                        beta = currentAlphaBeta;
                    }
                    //calculating branch value
                    if (currentScore <= bestScoreSoFar) {
                        if (currentScore < bestScoreSoFar) {
                           bestScoreSoFar=currentScore;
                           //System.out.println(i+" "+currentScore+" "+bestScoreSoFar);
                           if (i!=-1) {
                               bestVariationSoFar = allLegalMoves.get(i).toString()+" "+returnValue.getPrincipalVariation();
                           } else {
                               bestVariationSoFar = principalVariation[currentDepth].toString()+" "+returnValue.getPrincipalVariation();
                           }
                           if (currentDepth==0) {
                               bestMoves.clear();
                               if (SuicideChess.postThinkingOutput()) {
                                   System.out.println(maxDepth+extraSymbol+"\t"+returnValue.getBranchValue()+
                                           "\t"+((int)((new Date()).getTime()-thinkingBeginingTime.getTime())/10)+ //search time in centiseconds
                                           "\t"+nodesSearched+"\t"+bestVariationSoFar);
                               }
                               //System.out.println("*** Clear ");
                               if(bestScoreSoFar==Board.BLACK_WINS) { //found a win, no need to go further
                                   if(SuicideChess.playInACSII()) System.out.println("Found a win !");
                                   if (i!=-1) {
                                       bestMoves.add(allLegalMoves.get(i));
                                   } else {
                                       bestMoves.add(principalVariation[currentDepth]);
                                   }
                                   return new ReturnWrapper(beta,bestScoreSoFar,bestVariationSoFar);
                               }
                           }
                        } 
                        if(currentDepth==0) {
                            if(i!=-1) {
                                bestMoves.add(allLegalMoves.get(i));
                            } else {
                                bestMoves.add(principalVariation[currentDepth]);
                            }
                            //System.out.println("*** Adding "+allLegalMoves.get(i));
                        }
                    }
                    
                    if(beta<alpha) {
                        //if(currentDepth!=SuicideChess.getPlyDepth()-1) System.out.println("Pruning "+Integer.toString(allLegalMoves.size()-i)+" alternatives at depth "+ currentDepth);
                        return new ReturnWrapper(alpha,bestScoreSoFar,bestVariationSoFar); //pruning
                    }
                }
                return new ReturnWrapper(beta,bestScoreSoFar,bestVariationSoFar);
            } else { //white piece
                bestScoreSoFar=Board.MIN_VALUE; //white tries to maximise
                for (int i=-1; i<allLegalMoves.size(); i++) {
                    Board boardCopy = new Board(bitboard);
                    if (i==-1) { //do the principal variation first
                        if(!inPrincipalVariation || (principalVariation.length <= currentDepth)) {
                            continue;
                        } else {
                            boardCopy.doMove(principalVariation[currentDepth]);
                            //System.out.println(i+"wdoingMove"+currentDepth+" "+principalVariation[currentDepth]);
                            returnValue = AlphaBeta(boardCopy,currentDepth+1,true,currentMaxDepth,alpha,Board.MAX_VALUE);
                        }
                    } else { //and then don't do it again.
                        if(inPrincipalVariation && (principalVariation.length > currentDepth)) {
                            if(principalVariation[currentDepth].isSimpleEqualTo(allLegalMoves.get(i))) {
                                continue;
                            }
                        }
                        boardCopy.doMove(allLegalMoves.get(i));
                        //System.out.println(i+"wdoingMove"+currentDepth+" "+allLegalMoves.get(i)+" "+nodesSearched);
                        returnValue = AlphaBeta(boardCopy,currentDepth+1,false,currentMaxDepth,alpha,Board.MAX_VALUE);
                    }
                    
                    //System.out.println("Analysing "+currentDepth+":"+allLegalMoves.get(i));
                    
                    currentScore = returnValue.getBranchValue();
                    //if(currentDepth==0 || currentDepth==1) System.out.println(returnValue.getPrincipalVariation()+" "+currentScore);
                    currentAlphaBeta = returnValue.getAlphaBeta();
                    
                    //System.out.println("| CurrentScore, BestScore:" + currentScore + ", " + bestScoreSoFar);
                    //System.out.println("| CurrentAlpha, alpha:" + currentAlphaBeta + ", " + alpha);
                    
                    //calculating new value of alpha
                    if (currentAlphaBeta>alpha) {
                        alpha = currentAlphaBeta;
                    }
                    //calculating branch value

                    if (currentScore >= bestScoreSoFar) {
                        if (currentScore > bestScoreSoFar) {
                            bestScoreSoFar=currentScore;
                            //System.out.println(i+" "+currentScore+" "+bestScoreSoFar);
                            if (i!=-1) {
                                bestVariationSoFar = allLegalMoves.get(i).toString()+" "+returnValue.getPrincipalVariation();
                            } else {
                                bestVariationSoFar = principalVariation[currentDepth].toString()+" "+returnValue.getPrincipalVariation();
                            }
                            if (currentDepth==0) {
                                bestMoves.clear();
                                if (SuicideChess.postThinkingOutput()) {
                                    System.out.println(maxDepth+extraSymbol+"\t"+returnValue.getBranchValue()+
                                            "\t"+((int)((new Date()).getTime()-thinkingBeginingTime.getTime())/10)+ //search time in centiseconds
                                            "\t"+nodesSearched+"\t"+bestVariationSoFar);
                                }
                                if(bestScoreSoFar==Board.WHITE_WINS) { //found a win, no need to go further
                                    if(SuicideChess.playInACSII()) System.out.println("Found a win !");
                                    if (i!=-1) {
                                        bestMoves.add(allLegalMoves.get(i));
                                    } else {
                                        bestMoves.add(principalVariation[currentDepth]);
                                    }
                                    return new ReturnWrapper(alpha,bestScoreSoFar,bestVariationSoFar);
                                }
                                //System.out.println("*** Clear ");
                            }
                        }
                        if(currentDepth==0) {
                            if(i!=-1) {
                                bestMoves.add(allLegalMoves.get(i));
                            } else {
                                bestMoves.add(principalVariation[currentDepth]);
                            }
                        }
                    }

                    if(alpha>beta) {
                        //if(currentDepth!=SuicideChess.getPlyDepth()-1) System.out.println("Pruning "+Integer.toString(allLegalMoves.size()-i)+" alternatives at depth "+ currentDepth);
                        return new ReturnWrapper(beta,bestScoreSoFar,bestVariationSoFar); //pruning
                    }   
                }
                return new ReturnWrapper(alpha,bestScoreSoFar,bestVariationSoFar);
            }
        }  
    }
}