package suicideChess;

import suicideChess.Square.NotAValidSquare;

/**
 * This class is used for moves representation.
 * 
 * @author Jean-Baptiste H&eacute;tier
 * @version $LastChangedRevision$, $LastChangedDate$
 */

public class Move {

	/*======*
	 * DATA *
	 *======*/
	
	//integers 
	private Piece movingPiece;

	private Square fromSquare;
	private Square toSquare;
	private Piece promotionPiece;
	private boolean isPromotion;
	
	private boolean isCapture;
	private Piece capturePiece;
	
	//the following are used for two things :
    //moves that enable the 'en Passant' status of the board
    //and 'en passant' captures.
    private boolean isEnPassant=false;
	private Square enPassantSquare;

    @SuppressWarnings("serial")
	public class NotAValidMoveException extends Exception {
		NotAValidMoveException(String s) { super(s); };
	}
    
	/*=============*
	 * CONSTRUCTOR *
	 *=============*/

	/**
	 * This is the constructor of the class.
	 * 
	 * @param move A String containing of type "e2e4" (4 chars), "b7b8q" (5 chars) for promotions.
	 * It is the standard algebraic notation used for chess.
	 * @param board A {@link Board} to be able to locate the pieces.
	 * @throws NotAValidMoveException When the String is not legal (too long or too short).
	 * @throws NotAValidSquare It the String is not legal (not standard algebraic notation).
	 * @see Board
	 */
	public Move(String move, Board board) throws NotAValidMoveException, NotAValidSquare {
		isPromotion = false;
		
		switch (move.length()) {
			case 5:
				isPromotion = true;
                promotionPiece = new Piece(move.substring(4,5).toCharArray()[0],board.getCurrentPlayer());
				//no break statement here on purpose
			case 4:
				fromSquare = new Square(move.substring(0,2));
				toSquare = new Square(move.substring(2,4));
				break;
			default:
				throw new NotAValidMoveException("Invalid Move: "+move);
		}
		
		movingPiece = board.getPiece(fromSquare());
		if(movingPiece.getPieceNumber()==Piece.NONE) {
			throw new NotAValidMoveException("No piece to move on "+fromSquare.getFile()+fromSquare.getRank());
		}
		capturePiece = board.getPiece(toSquare());
		
		if (capturePiece.getPieceNumber()!=Piece.NONE) {
			isCapture = true;
		} else {
			isCapture = false;
		}
        
        if(movingPiece.getPieceType()==Piece.PAWN) {
            //check for enPassant capture
            if(board.isEnPassant() && (toSquare.isEqual(board.getEnPassantSquare()))) {
                isCapture = true;
                isEnPassant = true;
                if(movingPiece.getColor()==Piece.BLACK) { 
                    enPassantSquare = new Square(board.getEnPassantSquare().getFileNb(),board.getEnPassantSquare().getRank()+1);
                } else {
                    enPassantSquare = new Square(board.getEnPassantSquare().getFileNb(),board.getEnPassantSquare().getRank()-1);
                }
                capturePiece = board.getPiece(enPassantSquare);
            }
        
            //check for setting enPassant Square
            if(movingPiece.getColor()==Piece.BLACK && toSquare.getRank()==fromSquare.getRank()-2) {
                isEnPassant = true;
                enPassantSquare = new Square(fromSquare.getFileNb(),fromSquare.getRank()-1);
            }
            if(movingPiece.getColor()==Piece.WHITE && toSquare.getRank()==fromSquare.getRank()+2) {
                isEnPassant = true;
                enPassantSquare = new Square(fromSquare.getFileNb(),fromSquare.getRank()+1);
            }
        }
	}
	
	/**
	 * This is the constructor of the class.
	 * 
	 * @param fromSquare The {@link Square} to move from.
	 * @param toSquare The {@link Square} to move to.
	 * @param movingPiece The {@link Piece} to be moved.
	 * @param capturePiece The {@link Piece} to be captured. Piece.NONE if none.
	 * @param promotionPiece The {link Piece} to be promoted to. Piece.NONE if none.
	 * @throws NotAValidSquare If Squares are not valid.
	 * @see Square
	 * @see Piece
	 */
	public Move(Square fromSquare, Square toSquare, Piece movingPiece, Piece capturePiece, Piece promotionPiece)
			throws NotAValidSquare {

		this.fromSquare = fromSquare;
		this.toSquare = toSquare;
		
		this.movingPiece = movingPiece;

		this.capturePiece = capturePiece;
		if (capturePiece.getPieceNumber()!=Piece.NONE) {
			isCapture = true;
		} else {
			isCapture = false;
		}

		this.promotionPiece = promotionPiece;
		if (promotionPiece.getPieceNumber()!=Piece.NONE) {
			isPromotion = true;
		} else {
			isPromotion = false;
		}
	}


	/**
	 * This is the constructor of the class.
	 * It should be used only for moves that enable 'en passant' or are 'en passant' capture.
	 * 
	 * @param fromSquare The {@link Square} to move from.
	 * @param toSquare The {@link Square} to move to.
	 * @param enPassantSquare The 'en passant' {@link Square}.
	 * @param movingPiece The {@link Piece} to be moved.
     * @param isCapture This boolean is used to differenciate 'en passant' captures from moves that simply enable 'en passant' capture.
	 * @throws NotAValidSquare If Squares are not valid.
	 * @see Square
	 * @see Piece
	 */
	public Move(Square fromSquare, Square toSquare, Square enPassantSquare, Piece movingPiece, boolean isCapture)
			throws NotAValidSquare {

		this.fromSquare = fromSquare;
		this.toSquare = toSquare;
		
		this.movingPiece = movingPiece;

		this.isCapture = isCapture;
        if (isCapture) {
            if (movingPiece.getColor()==Piece.WHITE) {
                this.capturePiece=new Piece(Piece.BLACK_PAWN);
            } else {
                this.capturePiece=new Piece(Piece.WHITE_PAWN); 
            }
        }
            
		isPromotion = false;
		isEnPassant = true;
		this.enPassantSquare = enPassantSquare;
	}
	
	
	/*===========================================*
	 * Those two constructors are no longer used *
	 *===========================================*/
	/*
	//non capture move
	public Move(String move, Piece pieceToMove) throws NotAValidMoveException, NotAValidSquare {
		movingPiece = pieceToMove;
		isPromotion = false;
		isCapture = false;
		
		switch (move.length()) {
			case 5:
				isPromotion = true;
				promotionPiece = new Piece(move.toCharArray()[4], movingPiece.getColor());
				//no break statement here on purpose
			case 4:
				fromSquare = new Square(move.substring(0,2));
				toSquare = new Square(move.substring(2,4));
				break;
			default:
				throw new NotAValidMoveException("Invalid Move: "+move);
		}
	}
	
	//capture move
	public Move(String move, Piece pieceToMove, Piece pieceToCapture) throws NotAValidMoveException, NotAValidSquare {
		movingPiece = pieceToMove;
		isPromotion = false;
		isCapture = true;
		capturePiece = pieceToCapture;
		
		switch (move.length()) {
			case 5:
				isPromotion = true;
				promotionPiece = new Piece(move.toCharArray()[4], movingPiece.getColor());
				//no break statement here on purpose
			case 4:
				fromSquare = new Square(move.substring(0,2));
				toSquare = new Square(move.substring(2,4));
				break;
			default:
				throw new NotAValidMoveException("Invalid Move: "+move);
		}
	}

	*/
	
	
	/*=========*
	 * METHODS *
	 *=========*/
	
	/**
	 * Returns the {@link Square} to move from in a Move
	 * @return Square
	 * @see Square 
	 */
	public Square fromSquare() {
		return fromSquare;
	}

	/**
	 * Returns the {@link Square} to move to in a Move
	 * @return Square
	 * @see Square 
	 */
	public Square toSquare() {
		return toSquare;
	}
	
	/**
	 * Returns a boolean saying if a Move is a promotion or not.
	 * @return boolean
	 */
	public boolean isPromotionMove() {
		return isPromotion;
	}
	
	/**
	 * Returns the promotion {@link Piece} of a Move
	 * @return Piece
	 * @see Piece
	 */
	public Piece getPromotionPiece() {
		if (isPromotion) {
			return promotionPiece;
		} else {
			return new Piece(Piece.NONE);
		}
	}
	
	/**
	 * Returns the moving {@link Piece} of a Move.
	 * @return Piece
	 * @see Piece
	 */
	public Piece getMovingPiece() {
		return movingPiece;
	}
	
	/** 
	 * Returns a boolean saying if a Move is a capture move
	 * @return boolean
	 */
	public boolean isCaptureMove() {
		return isCapture;
	}
	
	/**
	 * Returns the {@link Piece} to be captured in a Move 
	 * @return Piece
	 */
	public Piece getCapturedPiece() {
		return capturePiece;
	}
	
	/**
	 * Returns a boolean saying is a Move makes 'en passant' move possible.
	 */
	public boolean enablesEnPassant() {
		return (!isCapture)&&isEnPassant;
	}
	
    /**
     * Returns a boolean saying is a Move is an 'en passant' move.
     */
    public boolean isEnPassant() {
        return (isCapture)&&isEnPassant;
    }
    
    /**
	 * Returns the enPassant Square for a move.
	 */
    public Square getEnPassantSquare() {
        return enPassantSquare;
    }
	
    
    /**
	 * Converts a Move into a String.
	 * @return String 
	 */
	public String toString() {
		if (isPromotion) {
            return fromSquare.toString()+toSquare+promotionPiece.toString();
        }
        return fromSquare.toString()+toSquare;
	}

	
    /**
     * Tests the equality of two moves by just looking at the fromSquare and the toSquare
     * @param that A move to test
     * @return boolean
     */
    public boolean isSimpleEqualTo(Move that) {
        return ((this.fromSquare.isEqual(that.fromSquare))&&(this.toSquare.isEqual(that.toSquare)));
    }
    
    
    /**
	 * Displays a move in great details.
	 */
	public void display() {
		System.out.println(" ");
		System.out.println("===== Move =====");
		System.out.println("Move:          "+getMovingPiece().toString());
		System.out.println("From square:  "+fromSquare());
		System.out.println("To square:    "+toSquare());
		System.out.println((isPromotionMove() ? "Promotion:     "+getPromotionPiece().toString() : "No Promotion"));
		System.out.println((isCaptureMove() ? "Capture:       "+getCapturedPiece().toString() : "No Capture"));
        if (isEnPassant()) System.out.println("En passant:    "+getEnPassantSquare().getFile()+getEnPassantSquare().getRank());
        System.out.println((enablesEnPassant() ? "Set en-pass.: "+getEnPassantSquare().getFile()+getEnPassantSquare().getRank() : "No en-pass. set"));
		System.out.println("================");
		System.out.println(" ");
	}
	
	
	/*======================================*
	 * I do not use those functions anymore *
	 *======================================*
	
	public void setFromSquare(Square square) {
		fromSquare = square;
	}
	public void setToSquare(Square square) {
		toSquare = square;
	}
	public void setPromotionMove(String piece) {
		if (piece.length()==0) {
			isPromotion=false;
		} else {
			isPromotion=true;
			promotionPiece=new Piece(piece);
		}
	}
	public void setMovingPiece(Piece piece) {
		movingPiece = piece;
	}
		
	*/
}
